/* cp8299x assembler  version 1.0

a3   7 September 1999

modify from a3  13 July 2004
modify from a4  17 December 2005

Prabhas Chongstitvatana

format  (case insensitive)

.org 0                    set address
label: value
 ...

label: opcode operand
 ...
.end                      end

operand = #sym  sym (sym)    immediate, direct, indirect
sym = numeric symbol label
numeric begins with 0..9 decimal/hex
symbol  found in the symbol table

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define	MXBUF	1024		/* max input line length */
#define	MXOP	16			/* number of opcode */
#define MXMEM	1000        /* size of memory */

char	lbuf[MXBUF];		   /* for listing */
char 	ibuf[MXBUF], *cp = NULL;   /* input buffer, ch ptr */
static 	char sep[] = " \t\n";      /* separator char */
char    *w;
int 	lineno,loc;		   /* line num, current ads */
FILE 	*fi;
char 	mem[MXMEM];

typedef unsigned char uchar;
typedef struct { char name[5]; char opn; } optype;

// opn: 0 IH, 1 IM+DI, 2 DI, 3 RE+DI
optype optable[] = {
	"LDS",1, "LDA",1, "PSH",0, "STA",2, "JMP",3, "JPZ",3,
	"JPC",3, "JSR",3, "AND",1, "ORA",1, "POP",0, "XOR",1,
	"ADD",1, "RTS",0, "ROL",0, "ROR",0
};

void putsym( char *name, int value);
int getsym( char *name);
void Swriter(char *mem, int a, int b);

void error(char *s){
  printf("line %d error : %s symbol %s\n%s",lineno,s,w,lbuf);
  exit(0);
}
// get one token from input
char *gettok(void){
  if( cp != NULL ) cp = strtok(NULL,sep);
  while ( cp == NULL ) {
    if( fgets(ibuf,MXBUF,fi) == NULL ) return NULL;
    lineno++;
    strcpy(lbuf,ibuf);
    cp = strtok(ibuf, sep);
  }
  return cp;
}
// get token, skip comment
char *tok(void){
  cp = gettok();
  while(cp[0] == ';'){	// skip comment
	  cp = NULL;
	  cp = gettok();
  }
  cp = strupr(cp);   // convert to uppercase
  return cp;
}

// size of operand
int sizeopr( char *w, int op ){
  switch( w[0] ) {
  case '#' : return op == 0 ? 2 : 1;	// LDS
//  case '@' : return 1;
  case '(' : return 2;
  default  : return 2;
  }
}

char last(char *s){	return s[strlen(s)-1]; }

// convert hex to integer
int htoi(char *s){
	char c;
	int v, h;
	h = 0;
	v = 0;
	while( (c = *s) != 'H' ){
		if(c >= '0' && c <= '9') v = (int)(c-'0');
		else if(c >= 'A' && c <= 'F') v = (int)(c-'A'+10);
		h = h * 16 + v;
		s++;
	}
	return h;
}
// convert number to integer
int cvi(char *s){
	if (last(s) == 'H') return htoi(s);
	return atoi(s);
}

int getvalue( char *w ){
  char *s;
  int v;

  if( w[0] == '#' ){
	  s = w+1;
  }else if( w[0] == '(' ){
	  s = w+1;
	  w[strlen(w)-1] = 0;
  }else
  	  s = w;
  if( isdigit(s[0]) || s[0] == '-')
	  return cvi(s);
  if(last(w) == ':'){
	  w[strlen(w)-1] = 0;
	  v = getsym(w);
  }else
  	  v = getsym(s);
  if( v < 0 ) error("symbol not define");
  return v;
}

// 1 IM, 2 RE, 3 DI 4 IND
int getmode( char *w ){
  switch( w[0] ) {
  case '#' : return 1;
//  case '@' : return 2;
  case '(' : return 4;
  default  : return 3;
  }
}

void put1( int v ){
  mem[loc++] = v & 255;
}

void put2( int v ){
  mem[loc++] = (v >> 8) & 255;
  mem[loc++] = v & 255;
}

int isop0( int c ){
  return(  optable[c].opn == 0 );
}

//   opn: 0 IH, 1 IM+DI, 2 DI, 3 RE+DI
//   mode: 0 IH, 1 IM, 2 RE, 3 DI, 4 indirect
void checkmode( int op, int m){
	int opn;
	opn = optable[op].opn;
	if( m == 4 && (op == 1 || op == 3)) return;
	if( opn == 1 && (m == 1 || m == 3)) return;
	if( opn == 2 && (m == 3)) return;
	if( opn == 3 && (m == 2 || m == 3)) return;
	error("addressing mode");
}

// generate one line listing at pc
void dolisting(int pc){
  printf("%04X : %02X", pc, (uchar)mem[pc]); // pc : opcode
  switch ( loc - pc ) {                      // 2 opr8, 3 opr16
  case 1 : printf("       "); break;
  case 2 : printf(" %02X    ", (uchar)mem[pc+1]); break;
  case 3 : printf(" %04X  ",256*(uchar)mem[pc+1]+(uchar)mem[pc+2]);
  }
  printf("%s", lbuf);
}
// two-pass assembler
void twopass(int pass){
  int v, op, s, m, mbit, mark;
  w = tok();
  while( strcmp(w,".END") != 0 ) {
	if(strcmp(w,".ORG") == 0){
		w = tok();				// get address
		loc = cvi(w);
	}else if( last(w) == ':'){	// label
  	  if( pass == 1 ){			// first pass, store
		 w[strlen(w)-1] = 0;
		 putsym(w, loc);
	  }else
	     printf("               %s",lbuf);  // listing
	}else if( isdigit(w[0]) ){
		put1(cvi(w));
	}else{
  	  mark = loc;
  	  op = getsym(w);			// opcode
      if( op < 0 ) error("undefined op");
      if( isop0(op) ){
        put1(op << 4);			// op with no arg
      }else{
        w = tok();				// get arg
        m = getmode(w);
	    checkmode(op,m);
		mbit = m == 4 ? 2 : (m << 2);
		put1((op << 4) | mbit);	// put op
		s = sizeopr(w,op);
		v = 0;
        if(pass == 2) v = getvalue(w);
	    if(s == 1) put1(v); else put2(v);
      }
      if(pass == 2) dolisting(mark);
    }
	w = tok();			// get next token
  }
}

/* make two new file names from source, name.lst and name.s */
void makename( char *source, char *lst, char *hex ){
  int n;
  n = strcspn(source,".");
  strncpy(lst,source,n);
  strcpy(lst+n,".lst");
  strncpy(hex,source,n);
  strcpy(hex+n,".s");
}

int main(int argc, char *argv[]){
  char flst[80], fs[80];
  int i;
  if( argc < 2 ) {
    printf("usage : asm inputfile\n");
    exit(0);
  }
  // put reserved word in sym table
  for(i=0; i<MXOP; i++)
    putsym(optable[i].name,i);
  fi = fopen(argv[1],"r");
  lineno = 0;
  loc = 0;
  twopass(1);
//  dumpsymbol();
  rewind(fi);
  makename(argv[1], flst, fs);
  freopen(flst,"w",stdout);
  lineno = 0;
  loc = 0;
  twopass(2);
  fclose(fi);
  freopen(fs,"w",stdout);
  Swriter(mem,0,loc);
  Swriter(mem,0,0);
  return 0;
}
