/* S format converter

Prabhas Chongstitvatana 9th September 1999


  S format object file

S113011013B100102F03B7001008FF00117E01086F
S1090010000000040506D7
S9030000FC

S1xxyyyyddddd....ss

xx length of total data (byte) on this line (yy...ss)
yyyy starting address
dd   data
ss   checksum  (add all xx..ss)  must be FF (255)

the last line begins with S9

explanation of the above example :
line 1 :  length 13h (19 bytes from 0110...6F )
	  address 0110h
	  data  13 B1 00 10 ... 08
	  check sum  6F ( FF - (add 13...08 ) )
line 2 :  length 9h (10 bytes)
	  address 0010h
	  data 00 00 00 00 04 05 06
	  check sum D7
line 3 :  last line

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define LOCAL  static

LOCAL char xbuf[256], *ibp = NULL;
LOCAL int chksum;

static char hexstr[] = "0123456789ABCDEF";

LOCAL int hex2dec( char c ){
  char *p;
  p = strchr( hexstr, (int)(c) & 255 );
  return (int)(p - hexstr);
}

LOCAL char dec2hex( int n ){
  if( n < 0 || n > 15 ) return 0;
  return hexstr[n];
}

LOCAL void writehex2( int n ){
  chksum += n;
  *ibp = dec2hex((n >> 4) & 15);
  *(ibp+1) = dec2hex(n & 15);
  ibp += 2;
}

LOCAL int readhex2(void){	/* read two hex from input buffer */
  int v;
  v = 16*hex2dec(*ibp++) + hex2dec(*ibp++);
  chksum += v;
  return v;
}

extern FILE *fi;

int Sreader(char *mem){  /* return the highest ads */
  int i,len,ads,max;

  max = 0;
  while( fgets(xbuf,256,fi) != NULL )  {
    ibp = xbuf;
    if( xbuf[0] != 'S' || xbuf[1] != '1' ) break;
    ibp += 2;
    chksum = 0;
    len = readhex2();
    ads = 256*readhex2() + readhex2();
    for(i=0; i<len-3; i++)
      mem[ads++] = (char)readhex2();
    readhex2();
    if( (chksum & 255) != 255 ) {
      printf("%s\ncheck sum error\n",xbuf);
      exit(0);
    }
   if ( ads > max ) max = ads;
  }
  return max;
}
/* from a to b  call with 0,0 to end */
void Swriter(char *mem, int a, int b){
  int ads,e,i;

  if( a == 0 && b == 0 )
    puts("S9030000FC"); /* last line */
  ads = a;
  while( ads < b ) {
    e = ( b - ads > 16 ) ? 16 : b - ads ;
    strcpy(xbuf,"S1");
    ibp = xbuf + 2;
    chksum = 0;
    writehex2(e+3);
    writehex2( (ads & 0x0FF00) >> 8);
    writehex2( ads & 255 );
    for(i=ads; i<ads+e; i++ )
      writehex2( mem[i] ) ;
    writehex2( 255-(chksum & 255) );
    ads += e;
    *ibp = 0;
    puts(xbuf);
  }
}

/* test with max.obj
S1130100FE0011A613B70010B6001281022E11A62C
S113011013B100102F03B7001008FF00117E01086F
S1090010000000040506D7
S9030000FC
*/
/*
main()
{
  int loc;
  char M[1000];

  loc = Sreader(M);
  Swriter(M,256,288);
  Swriter(M,16,22);
  Swriter(M,0,0);
}
*/
